"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
from q3dviewer.utils.maths import make_transform

def get_dtype(msg):
    dtype_map = {
        1: 'i1',   # INT8
        2: 'u1',   # UINT8  
        3: 'i2',   # INT16
        4: 'u2',   # UINT16
        5: 'i4',   # INT32
        6: 'u4',   # UINT32
        7: 'f4',   # FLOAT32
        8: 'f8'    # FLOAT64
    }
    
    point_step = msg.point_step
    dtype_list = []
    for field in msg.fields:
        dtype_str = dtype_map.get(field.datatype, 'f4')
        dtype_list.append((field.name, dtype_str, field.offset))
    
    dtype_list.sort(key=lambda x: x[2])
    
    structured_dtype = []
    last_offset = 0
    
    for field_name, field_dtype, offset in dtype_list:
        if offset > last_offset:
            padding_size = offset - last_offset
            if padding_size > 0:
                structured_dtype.append((f'pad{last_offset}', f'u1', padding_size))
        
        structured_dtype.append((field_name, field_dtype))
        last_offset = offset + np.dtype(field_dtype).itemsize
    
    if point_step > last_offset:
        final_padding = point_step - last_offset
        structured_dtype.append((f'pad{last_offset}', f'u1', final_padding))
    return structured_dtype


def convert_pointcloud2_msg(msg):
    # Build dtype with proper offsets for PointCloud2 message
    structured_dtype = get_dtype(msg)
    pc = np.frombuffer(msg.data, dtype=structured_dtype)
    # pc = PointCloud.from_msg(msg).pc_data
    data_type = [('xyz', '<f4', (3,)), ('irgb', '<u4')]
    rgb = np.zeros([pc.shape[0]], dtype=np.uint32)
    intensity = np.zeros([pc.shape[0]], dtype=np.uint32)
    fields = ['xyz']
    if 'intensity' in pc.dtype.names:
        intensity = pc['intensity']
        intensity = np.clip(intensity, 0, 255)
        intensity = intensity.astype(np.uint32)
        fields.append('intensity')
    if 'rgb' in pc.dtype.names:
        rgb = pc['rgb'].view(np.uint32)
        fields.append('rgb')
    irgb = (intensity << 24) | rgb
    xyz = np.stack([pc['x'], pc['y'], pc['z']], axis=1)
    cloud = np.rec.fromarrays([xyz, irgb], dtype=data_type)
    stamp = msg.header.stamp.to_sec()
    return cloud, fields, stamp


def convert_odometry_msg(msg):
    pose = np.array(
        [msg.pose.pose.position.x,
         msg.pose.pose.position.y,
         msg.pose.pose.position.z])
    rotation = np.array([
        msg.pose.pose.orientation.x,
        msg.pose.pose.orientation.y,
        msg.pose.pose.orientation.z,
        msg.pose.pose.orientation.w])
    transform = make_transform(pose, rotation)
    stamp = msg.header.stamp.to_sec()
    return transform, stamp


def convert_image_msg(msg, bgr=False):
    image = np.frombuffer(msg.data, dtype=np.uint8).reshape(
        msg.height, msg.width, -1)
    # if bgr is True return BGR image, else return RGB image
    if (bgr and msg.encoding == 'rgb8') or (not bgr and msg.encoding == 'bgr8'):
        image = image[:, :, ::-1]

    stamp = msg.header.stamp.to_sec()
    return image, stamp



"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

from q3dviewer.custom_items.cloud_item import CloudItem
from pathlib import Path
import os
from q3dviewer.Qt.QtWidgets import QPushButton, QLabel, QLineEdit, QMessageBox
from q3dviewer.utils.cloud_io import save_pcd, save_ply, save_e57, save_las, load_pcd, load_ply, load_e57, load_las

class CloudIOItem(CloudItem):
    """
    A OpenGL point cloud item with input/output capabilities.
    Attributes:
        size (float): The size of each point. When `point_type` is 'PIXEL', this is the size in screen pixels.
            When `point_type` is 'SQUARE' or 'SPHERE', this is the size in centimeters in 3D space.
        alpha (float): The transparency of the points, in the range [0, 1], where 0 is fully transparent and 1 is fully opaque.
        color_mode (str): The coloring mode for the points.
            - 'FLAT': Single flat color for all points (uses the `color` attribute).
            - 'I': Color by intensity.
            - 'RGB': Per-point RGB color.
            - 'GRAY': Per-point grayscale color.
        color (str or tuple): The flat color to use when `color_mode` is 'FLAT'. Accepts any valid matplotlib color (e.g., 'red', '#FF4500', (1.0, 0.5, 0.0)).
        point_type (str): The type/rendering style of each point:
            - 'PIXEL': Draw each point as a square pixel on the screen.
            - 'SQUARE': Draw each point as a square in 3D space.
            - 'SPHERE': Draw each point as a sphere in 3D space.
        depth_test (bool): Whether to enable depth testing. If True, points closer to the camera will appear in front of farther ones.
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.save_path = str(Path(os.path.expanduser("~"), "data.pcd"))

    def add_setting(self, layout):
        super().add_setting(layout)

        label4 = QLabel("Save Path:")
        layout.addWidget(label4)
        box4 = QLineEdit()
        box4.setText(self.save_path)
        box4.textChanged.connect(self.set_path)
        layout.addWidget(box4)
        save_button = QPushButton("Save Cloud")
        save_button.clicked.connect(self.save)
        layout.addWidget(save_button)
        self.save_msg = QMessageBox()
        self.save_msg.setIcon(QMessageBox.Information)
        self.save_msg.setWindowTitle("save")
        self.save_msg.setStandardButtons(QMessageBox.Ok)

    def save(self):
        cloud = self.buff[:self.valid_buff_top]
        func = None
        if self.save_path.endswith(".pcd"):
            from q3dviewer.utils.cloud_io import save_pcd
            func = save_pcd
        elif self.save_path.endswith(".ply"):
            from q3dviewer.utils.cloud_io import save_ply
            func = save_ply
        elif self.save_path.endswith(".e57"):
            from q3dviewer.utils.cloud_io import save_e57
            func = save_e57
        elif self.save_path.endswith(".las"):
            from q3dviewer.utils.cloud_io import save_las
            func = save_las
        elif self.save_path.endswith(".tif") or self.save_path.endswith(".tiff"):
            print("Do not support save as tif type!")
        else:
            print("Unsupported cloud file type!")
        try:
            func(cloud, self.save_path)
            self.save_msg.setText("Save cloud to  %s" % self.save_path)
        except Exception as e:
            print(e)
            self.save_msg.setText("Cannot save to %s" % self.save_path)
            self.save_msg.exec()
        self.save_msg.exec()

    def load(self, file, append=False):
        # print("Try to load %s ..." % file)
        if file.endswith(".pcd"):
            from q3dviewer.utils.cloud_io import load_pcd
            cloud = load_pcd(file)
        elif file.endswith(".ply"):
            from q3dviewer.utils.cloud_io import load_ply
            cloud = load_ply(file)
        elif file.endswith(".e57"):
            from q3dviewer.utils.cloud_io import load_e57
            cloud = load_e57(file)
        elif file.endswith(".las"):
            from q3dviewer.utils.cloud_io import load_las
            cloud = load_las(file)
        else:
            print("Not supported file type.")
            return
        self.set_data(data=cloud, append=append)

        has_intensity = cloud['irgb'][0] & 0xff000000 > 0
        has_rgb = cloud['irgb'][0] & 0x00ffffff > 0

        if has_rgb:
            self.set_color_mode('RGB')
        elif has_intensity:
            self.set_color_mode('I')
        else:
            self.set_color_mode('FLAT')

        return cloud

    def set_path(self, path):
        self.save_path = path



#!/usr/bin/env python3

"""
Copyright 2024 Panasonic Advanced Technology Development Co.,Ltd. (Liu Yang)
Distributed under MIT license. See LICENSE for more information.
"""

import numpy as np
import q3dviewer as q3d
from q3dviewer.Qt.QtWidgets import QVBoxLayout, QProgressBar, QDialog, QLabel
from q3dviewer.Qt.QtCore import QThread, Signal, Qt
from q3dviewer.Qt.QtGui import QKeyEvent
from q3dviewer import GLWidget


class ProgressDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Loading Cloud File")
        self.setModal(True)
        self.progress_bar = QProgressBar(self)
        self.file_label = QLabel(self)
        layout = QVBoxLayout()
        layout.addWidget(self.file_label)
        layout.addWidget(self.progress_bar)
        self.setLayout(layout)

    def set_value(self, value):
        self.progress_bar.setValue(value)

    def set_file_name(self, file_name):
        self.file_label.setText(f"Loading: {file_name}")

    def closeEvent(self, event):
        if self.parent().progress_thread and self.parent().progress_thread.isRunning():
            event.ignore()
        else:
            event.accept()


class FileLoaderThread(QThread):
    progress = Signal(int)
    finished = Signal()

    def __init__(self, viewer, files):
        super().__init__()
        self.viewer = viewer
        self.files = files

    def run(self):
        cloud_item = self.viewer['cloud']
        for i, url in enumerate(self.files):
            file_path = url.toLocalFile()
            self.viewer.progress_dialog.set_file_name(file_path)
            cloud = cloud_item.load(file_path, append=(i > 0))
            center = np.nanmean(cloud['xyz'].astype(np.float64), axis=0)
            self.viewer.glwidget.set_cam_position(center=center)
            self.progress.emit(int((i + 1) / len(self.files) * 100))
        self.finished.emit()


class CustomGLWidget(GLWidget):
    def __init__(self, viewer):
        super().__init__()
        self.viewer = viewer
        self.selected_points = []

    def mousePressEvent(self, event):
        # ctrl + left click to add select point
        # ctrl + right click to remove a point from selected points
        if event.button() == Qt.LeftButton and event.modifiers() & Qt.ControlModifier:
            x, y = event.x(), event.y()
            p = self.get_point(x, y)
            if p is not None:
                self.selected_points.append(p)
                self.update_marker()
        elif event.button() == Qt.RightButton and event.modifiers() & Qt.ControlModifier:
            if len(self.selected_points) > 0:
                self.selected_points.pop()
                self.update_marker()
        super().mouseReleaseEvent(event)

    def update_marker(self):
        marks = []
        for p in self.selected_points:
            m = {'text': '',
                 'position': p,
                 'color': (0.0, 1.0, 0.0, 1.0),
                 'font_size': 16,
                 'point_size': 5.0,
                 'line_width': 1.0}
            marks.append(m)
        
        # calculate distances between consecutive points
        self.viewer['marker'].set_data(data=marks, append=False)
        if len(self.selected_points) >= 2:
            total_dists = 0.0
            for i in range(1, len(self.selected_points)):
                p1 = self.selected_points[i - 1]
                p2 = self.selected_points[i]
                dist = np.linalg.norm(np.array(p2) - np.array(p1))
                total_dists += dist
            self.viewer['text'].set_data(text=f'Total Distance: {total_dists:.2f} meter')
        else:
            self.viewer['text'].set_data(text='')


class CloudViewer(q3d.Viewer):
    def __init__(self, **kwargs):
        gl_widget_class=lambda: CustomGLWidget(self)
        super(CloudViewer, self).__init__(
            **kwargs,  gl_widget_class=gl_widget_class)
        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        """
        Overwrite the drop event to open the cloud file.
        """
        self.progress_dialog = ProgressDialog(self)
        self.progress_dialog.show()
        files = event.mimeData().urls()
        self.progress_thread = FileLoaderThread(self, files)
        self['cloud'].load(files[0].toLocalFile(), append=False)
        self.progress_thread.progress.connect(self.file_loading_progress)
        self.progress_thread.finished.connect(self.file_loading_finished)
        self.progress_thread.start()

    def file_loading_progress(self, value):
        self.progress_dialog.set_value(value)

    def file_loading_finished(self):
        self.progress_dialog.close()

    def open_cloud_file(self, file, append=False):
        cloud_item = self['cloud']
        if cloud_item is None:
            print("Can't find clouditem.")
            return
        cloud = cloud_item.load(file, append=append)
        center = np.nanmean(cloud['xyz'].astype(np.float64), axis=0)
        self.glwidget.set_cam_position(center=center)

# print a quick help message
def print_help():
    # ANSI color codes
    GREEN = '\033[92m'
    BLUE = '\033[94m'
    BOLD = '\033[1m'
    END = '\033[0m'
    
    help_msg = f"""
{BOLD}Cloud Viewer Help:{END}
{GREEN}• Drag and drop cloud files into the viewer to load them.{END}
{GREEN}• Measure distance between points:{END}
    {BLUE}- Hold Ctrl and left-click to select points on the cloud.{END}
    {BLUE}- Hold Ctrl and right-click to remove the last selected point.{END}
    {BLUE}- The total distance between selected points will be displayed.{END}
{GREEN}• Press 'M' to open the settings window.{END}
    {BLUE}- Use the settings window to adjust item properties.{END}
    """
    print(help_msg)

def main():
    print_help()
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--path", help="the cloud file path")
    args = parser.parse_args()
    app = q3d.QApplication(['Cloud Viewer'])
    viewer = CloudViewer(name='Cloud Viewer')
    cloud_item = q3d.CloudIOItem(size=1, alpha=0.1)
    axis_item = q3d.AxisItem(size=0.5, width=5)
    grid_item = q3d.GridItem(size=1000, spacing=20)
    marker_item = q3d.Text3DItem()  # Changed from CloudItem to Text3DItem
    text_item = q3d.Text2DItem(pos=(20, 40), text="", color='lime', size=16)

    viewer.add_items(
        {'marker': marker_item, 
         'cloud': cloud_item, 
         'grid': grid_item, 
         'axis': axis_item, 
         'text': text_item})

    if args.path:
        pcd_fn = args.path
        viewer.open_cloud_file(pcd_fn)

    viewer.show()
    app.exec()


if __name__ == '__main__':
    main()



from q3dviewer.custom_items.axis_item import AxisItem
from q3dviewer.custom_items.cloud_item import CloudItem
from q3dviewer.custom_items.cloud_io_item import CloudIOItem
from q3dviewer.custom_items.gaussian_item import GaussianItem
from q3dviewer.custom_items.frame_item import FrameItem
from q3dviewer.custom_items.grid_item import GridItem
from q3dviewer.custom_items.text_item import Text2DItem
from q3dviewer.custom_items.image_item import ImageItem
from q3dviewer.custom_items.line_item import LineItem
from q3dviewer.custom_items.text3d_item import Text3DItem



import time
t1 = time.time()
from q3dviewer.custom_items import *
t2 = time.time()
from q3dviewer.glwidget import *
from q3dviewer.viewer import *
from q3dviewer.base_item import *
from q3dviewer.base_glwidget import *
t3 = time.time()

from q3dviewer.Qt import Q3D_DEBUG
if Q3D_DEBUG:
    print("Import custom items: ", t2 - t1)
    print("Import base q3dviewer: ", t3 - t2)
